package info.textgrid.middleware;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigInteger;
import java.text.ParseException;
import java.util.Hashtable;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.classicmayan.tools.ImageMetsMods;
import org.classicmayan.tools.MediaHarvester;
import org.classicmayan.tools.TextGridUri;
import org.json.JSONArray;
import org.json.JSONException;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import info.textgrid.clients.tgcrud.CrudClientException;
import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.HeaderType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.ListRecordsType;
import info.textgrid.middleware.oaipmh.MetadataType;
import info.textgrid.middleware.oaipmh.RecordType;
import info.textgrid.middleware.oaipmh.ResumptionTokenType;

/**
 *
 */
public class IDIOMImages implements RecordDelivererInterface, RecordListDelivererInterface,
    IdentifierListDelivererInterface {

  private static Logger log = Logger.getLogger(IDIOMImages.class.getName());

  private String kindID;
  private String perPage;
  private String pageNumber;
  private int totalHits;

  protected static Map<String, Integer> cursorCollector = new Hashtable<String, Integer>();
  protected static Map<String, String> fromUntilCollector = new Hashtable<String, String>();

  /**
   * 
   */
  public IDIOMImages() {
    //
  }

  /**
   * 
   * @throws JSONException
   * @throws IOException
   * @throws ParseException
   * @throws SAXException
   * @throws TransformerException
   * @throws TransformerFactoryConfigurationError
   */

  private String from;
  private String until;

  /**
   *
   */
  @Override
  public ListRecordsType getRecords(final String from, final String to, final String set,
      final String resumptionToken)
      throws JSONException, IOException, ParseException, TransformerFactoryConfigurationError {

    String useThisFrom = from;
    String useThisTo = to;

    if (resumptionToken != null) {
      String fromUntil = fromUntilCollector.get(resumptionToken);
      useThisFrom = fromUntil.split(":")[0];
      useThisTo = fromUntil.split(":")[1];
    }

    MediaHarvester mediaHarvester =
        new MediaHarvester(getKindID(), getPerPage(), getPageNumber(), useThisFrom, useThisTo);
    JSONArray mediaList = mediaHarvester.getMediaListFromConedaKor().getJSONArray("records");
    ListRecordsType imageMetsModsList = new ListRecordsType();

    setTotalHits(mediaHarvester.getTotalMedia());
    this.setFrom(useThisFrom);
    this.setUntil(useThisTo);
    for (int i = 0; i < mediaList.length(); i++) {
      ImageMetsMods imm = new ImageMetsMods(mediaList.getJSONObject(i));
      RecordType record = new RecordType();
      MetadataType metadataMets = new MetadataType();

      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder;
      try {
        builder = factory.newDocumentBuilder();

        Document doc = builder.parse(new InputSource(new StringReader(imm.getXML())));
        metadataMets.setAny(doc.getDocumentElement());
        record.setMetadata(metadataMets);
        String dateStamp = mediaList.getJSONObject(i).get("created_at").toString();
        String identifier = mediaList.getJSONObject(i).get("id").toString();

        record.setHeader(buildOAIPMH_RecordHeader(dateStamp, identifier));
        imageMetsModsList.getRecord().add(record);
      } catch (ParserConfigurationException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      } catch (SAXException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    }
    imageMetsModsList.setResumptionToken(
        buildResumptionToken(resumptionToken, getTotalHits(), useThisFrom, useThisTo));

    return imageMetsModsList;
  }

  /**
   *
   */
  @Override
  public ListIdentifiersType processIdentifierList(final String from, final String to,
      final String set, final String resumptionToken)
      throws JSONException, IOException, ParseException {

    String useThisFrom = from;
    String useThisTo = to;

    if (resumptionToken != null) {
      String fromUntil = fromUntilCollector.get(resumptionToken);
      useThisFrom = fromUntil.split(":")[0];
      useThisTo = fromUntil.split(":")[1];
    }
    MediaHarvester mediaHarvester =
        new MediaHarvester(getKindID(), getPerPage(), getPageNumber(), useThisFrom, useThisTo);
    JSONArray mediaList = mediaHarvester.getMediaListFromConedaKor().getJSONArray("records");
    ListIdentifiersType identifiersList = new ListIdentifiersType();

    setTotalHits(mediaHarvester.getTotalMedia());

    for (int i = 0; i < mediaList.length(); i++) {
      HeaderType header = new HeaderType();
      header.setDatestamp(mediaList.getJSONObject(i).get("created_at").toString());
      header.setIdentifier(mediaList.getJSONObject(i).get("id").toString());
      identifiersList.getHeader().add(header);
    }

    identifiersList.setResumptionToken(
        buildResumptionToken(resumptionToken, getTotalHits(), useThisFrom, useThisTo));
    return identifiersList;
  }

  /**
   *
   */
  @Override
  public GetRecordType getRecordById(String id) throws JSONException, IOException, ParseException {

    GetRecordType singleImageMetsMods = new GetRecordType();
    RecordType conedaKorRecord = new RecordType();
    MetadataType metadataMets = new MetadataType();

    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder;

    // ImageMetsMods imm = new ImageMetsMods(new ConedaKorID(id));
    ImageMetsMods immByTGURI = null;

    if (!id.startsWith("textgrid:")) {
      id = "textgrid:" + id;
    }

    log.fine("processing id: " + id);

    try {
      immByTGURI = new ImageMetsMods(new TextGridUri(id));

      log.fine("image id/title: " + immByTGURI.getID() + " / " + immByTGURI.getTitle());

      builder = factory.newDocumentBuilder();

      log.fine("immByTGURI: " + (immByTGURI != null ? immByTGURI.getXML() : "NULL"));

      Document doc = builder.parse(new InputSource(new StringReader(immByTGURI.getXML())));
      metadataMets.setAny(doc.getDocumentElement());
      conedaKorRecord.setMetadata(metadataMets);
      if (id.contains(".")) {
        conedaKorRecord.setHeader(buildOAIPMH_RecordHeader(
            OAIPMHUtilities.datestampAsString(immByTGURI.getCreationDate()),
            id.substring(0, id.indexOf("."))));
      } else {
        conedaKorRecord.setHeader(buildOAIPMH_RecordHeader(
            OAIPMHUtilities.datestampAsString(immByTGURI.getCreationDate()), id));
      }

      singleImageMetsMods.setRecord(conedaKorRecord);

    } catch (ParserConfigurationException | SAXException | CrudClientException e) {
      log.severe("error! " + e.getClass().getName() + ": " + e.getMessage());
      return null;
    }

    return singleImageMetsMods;
  }

  /*
   * public ListRecordsType getAllMedia() throws JSONException, IOException, ParseException,
   * SAXException, TransformerFactoryConfigurationError, TransformerException{ ListRecordsType
   * imageMetsModsList = new ListRecordsType(); MediaHarvester mediaHarvester = new
   * MediaHarvester(getKindID(),getPerPage(), getPageNumber()); int perPage = 100;
   * setTotalHits(mediaHarvester.getTotalMedia()); for(int j=1; j*perPage<=100 ;j++){ mediaHarvester
   * = new MediaHarvester(getKindID(),getPerPage(), Integer.toString(j)); JSONArray mediaList =
   * mediaHarvester.getMediaListFromConedaKor().getJSONArray("records"); int harvestedEntities =
   * j*perPage; DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
   * DocumentBuilder builder; for(int i=0; i<mediaList.length(); i++){ try{ ImageMetsMods imm = new
   * ImageMetsMods(mediaList.getJSONObject(i));
   * 
   * 
   * builder = factory.newDocumentBuilder(); Document doc = builder.parse(new InputSource(new
   * StringReader(imm.getXML())));
   * 
   * RecordType record = new RecordType(); MetadataType metadataMets = new MetadataType();
   * metadataMets.setAny(doc.getDocumentElement()); record.setMetadata(metadataMets);
   * imageMetsModsList.getRecord().add(record); //printXML(doc);
   * 
   * } catch (ParserConfigurationException e) { // TODO Auto-generated catch block
   * e.printStackTrace(); }
   * 
   * }
   * 
   * } return imageMetsModsList; }
   */

  /**
   * @param dateStamp
   * @param Identifier
   * @return
   */
  public HeaderType buildOAIPMH_RecordHeader(String dateStamp, String Identifier) {
    HeaderType recordHeader = new HeaderType();
    recordHeader.setDatestamp(dateStamp);
    recordHeader.setIdentifier(Identifier);
    return recordHeader;
  }

  /**
   * @param resumptionToken
   * @param completeListSize
   * @param from
   * @param until
   * @return
   */
  public ResumptionTokenType buildResumptionToken(String resumptionToken, int completeListSize,
      String from, String until) {
    ResumptionTokenType resTokenForResponse = new ResumptionTokenType();
    if (resumptionToken != null && cursorCollector.containsKey(resumptionToken)) {
      if (cursorCollector.get(resumptionToken) * 100 > completeListSize) {
        resTokenForResponse.setValue("");
      } else {
        resTokenForResponse.setValue(resumptionToken);
      }
      resTokenForResponse.setCompleteListSize(BigInteger.valueOf(completeListSize));
      resTokenForResponse.setCursor(BigInteger.valueOf(cursorCollector.get(resumptionToken) * 100));
      cursorCollector.put(resumptionToken, cursorCollector.get(resumptionToken) + 1);
    } else {
      UUID uuid = UUID.randomUUID();
      String resTokenValue = uuid.toString();
      resTokenForResponse.setCompleteListSize(BigInteger.valueOf(completeListSize));
      resTokenForResponse.setCursor(BigInteger.valueOf(100));
      resTokenForResponse.setValue(resTokenValue);
      cursorCollector.put(resTokenValue, 2);
      fromUntilCollector.put(resTokenValue, from + ":" + until);
    }
    return resTokenForResponse;
  }

  /**
   * @param doc
   * @return
   * @throws TransformerFactoryConfigurationError
   * @throws TransformerException
   */
  public String printXML(Document doc)
      throws TransformerFactoryConfigurationError, TransformerException {
    Transformer transformer = TransformerFactory.newInstance().newTransformer();
    transformer.setOutputProperty(OutputKeys.INDENT, "yes");
    transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
    // initialize StreamResult with File object to save to file
    StreamResult result = new StreamResult(new StringWriter());
    DOMSource source = new DOMSource(doc);
    transformer.transform(source, result);
    String xmlString = result.getWriter().toString();
    return xmlString;
  }

  // **
  // GETTERS & SETTERS
  // **

  /**
   * @return
   */
  public int getTotalHits() {
    return this.totalHits;
  }

  /**
   * @param totalHits
   */
  public void setTotalHits(int totalHits) {
    this.totalHits = totalHits;
  }

  /**
   * @return
   */
  public String getPageNumber() {
    return this.pageNumber;
  }

  /**
   * @param pageNumber
   */
  public void setPageNumber(String pageNumber) {
    this.pageNumber = pageNumber;
  }

  /**
   * @return
   */
  public String getPerPage() {
    return this.perPage;
  }

  /**
   * @param perPage
   */
  public void setPerPage(String perPage) {
    this.perPage = perPage;
  }

  /**
   * @return
   */
  public String getKindID() {
    return this.kindID;
  }

  /**
   * @param kindID
   */
  public void setKindID(String kindID) {
    this.kindID = kindID;
  }

  /**
   * @return
   */
  public String getFrom() {
    return this.from;
  }

  /**
   * @param from
   */
  public void setFrom(String from) {
    this.from = from;
  }

  /**
   * @return
   */
  public String getUntil() {
    return this.until;
  }

  /**
   * @param until
   */
  public void setUntil(String until) {
    this.until = until;
  }

}
